from __future__ import annotations

from typing import List, Union

import base64
import datetime
import io
import secrets
import urllib.parse
import webbrowser
import urllib
import json
import requests

import Cryptodome.PublicKey.DSA
import Cryptodome.Hash.SHA256
import Cryptodome.Signature.DSS

import logging
import re
import time
import threading
import deephaven_enterprise.proto.auth_service_pb2_grpc
import deephaven_enterprise.proto.auth_pb2
import grpc

import deephaven_enterprise.client.util


class RefreshThread(threading.Thread):
    def __init__(self, auth_client: "AuthClient", slack_seconds: int = 60 * 2):
        """
        Create a thread to refresh the authentication cookie.  If the cookie is not refreshed within the deadline,
        then we will become unauthenticated and not be able to generate any more tokens.
        :param auth_client: auth client to refresh
        :param slack_seconds: how long before the cookie's expiration, in seconds, should we begin the refresh.
        Defaults to two minutes.
        """
        threading.Thread.__init__(self)
        self.daemon = True
        self.auth_client = auth_client
        self.slack = slack_seconds

    def run(self):
        # We don't want to hit the server too hard
        minimum_sleeps: List[int] = [1, 1, 2, 3, 5, 8, 13]
        sleep_index = 0

        while self.auth_client.cookie is not None:
            start_time = time.time()
            next_deadline_millis = self.auth_client.cookie_deadline_time_millis
            if next_deadline_millis is None or next_deadline_millis < start_time:
                # we've failed to refresh, we could try re-authentication with a private key if that was our method
                break
            next_deadline_seconds = next_deadline_millis / 1000

            to_sleep: float = next_deadline_seconds - start_time - self.slack
            time.sleep(to_sleep)
            self.auth_client.ping()
            # We expect the deadline to handle most sleeps, this logic prevents tight loops.
            duration_seconds: float = (time.time() - start_time)
            if duration_seconds < minimum_sleeps[sleep_index]:
                catchup_seconds = minimum_sleeps[sleep_index] - \
                    duration_seconds
                time.sleep(catchup_seconds)
                sleep_index = min(len(minimum_sleeps) - 1, sleep_index + 1)
            else:
                sleep_index = 0


class AuthClient:
    """
    AuthClient authenticates to a Deephaven authentication server and produces tokens for use with other Deephaven
    services.

    Presently, password and private key authentication are provided.
    """
    cookie_deadline_time_millis: int
    refresh_thread: "RefreshThread"
    channel: grpc.Channel
    me: str
    opened_channel: bool
    rpc_timeout_secs: int
    log: logging.Logger

    def __init__(self, host: str = None, port: int = None, rpc_timeout_seconds: int = 120, channel: grpc.Channel = None, channel_options=None, client_name=None):
        """
        Create an AuthClient and connect to the server.

        You may either specify the host and port to connect to, or provide your own grpc channel.  The simplest case
        is to simply provide the host and port, but if you need advanced channel configuration or want to share the
        channel for several clients, then you can create it and pass it in.

        :param host: the host to connect to, requires port
        :param port: the port to connect to, requires host
        :param rpc_timeout_seconds: the rpc timeout period to use, defaults to 120 seconds if not provided
        :param channel: a pre-created channel to use for the gRPC messages, it will not be closed by this client.
        :param channel_options: a list of options for channel creation (not used if the channel is provided)
        :param client_name: a name to append to the client identifier, useful for debugging and logging.
        """
        self.cookie = None
        self.refresh_thread = None
        self.cookie_deadline_time_millis = None
        self.rpc_timeout_secs = rpc_timeout_seconds
        self.log = logging.getLogger("deephaven_enterprise.client.auth")
        (self.channel, self.me) = deephaven_enterprise.client.util.get_grpc_channel(
            channel=channel, host=host, port=port, channel_options=channel_options)
        if client_name is not None:
            self.me += '_' + client_name
        self.opened_channel = channel is None
        self.clientID = deephaven_enterprise.client.util.make_client_id(
            self.me)
        self.me += '_' + deephaven_enterprise.client.util._uuid_str(self.clientID.uuid)
        self.log.info(f'{self.me} starting')

        self.stub = deephaven_enterprise.proto.auth_service_pb2_grpc.AuthApiStub(
            self.channel)

        self._send_ping()

    def _send_ping(self) -> float:
        """
        Sends a common ping request over the stub.
        :param stub: the stub to send a ping request over
        :param me: client identifier
        :return: the RTT in milliseconds of the ping
        """
        pingreq = deephaven_enterprise.proto.common_pb2.PingRequest()
        sendTime = time.time_ns()
        pingreq.sender_send_time_millis = int(sendTime / 1_000_000)
        pingreq.__setattr__("from", self.me)
        response = self.stub.ping(pingreq, timeout=self.rpc_timeout_secs, wait_for_ready=True)
        rtt = (time.time_ns() - sendTime) / 1_000_000
        server_time = datetime.datetime.utcfromtimestamp(response.receiver_receive_time_millis/1_000)
        server = getattr(response, 'from')
        self.log.info(
            f'{self.me} sent ping RPC request and got response, rountrip={rtt:.3f}s, ' +
            
            f'answer from={server} at {server_time}')
        return rtt

    def external_login(self, auth_key: Union[str, bytes]) -> None:
        """
        Authenticate to the server using an external authentication method, such as SAML or a custom, non-standard external
        authentication module.

        This requires that the server be configured with the external authentication module and that the external
        authentication module should be configured to accept the provided auth_key.

        :param auth_key: the authentication key used for external login.  This must be a string or bytes.
        """
        abe: deephaven_enterprise.proto.auth_pb2.AuthenticateByExternalRequest = deephaven_enterprise.proto.auth_pb2.AuthenticateByExternalRequest()
        abe.client_id.CopyFrom(self.clientID)
        if isinstance(auth_key, bytes):
            # If auth_key is bytes, decode it to a string
            auth_key = auth_key.decode('utf-8')
        abe.key = auth_key

        start_time = time.time()
        ar: deephaven_enterprise.proto.auth_pb2.AuthenticateByExternalLoginResponse = self.stub.authenticateByExternal(
            abe)
        end_time = time.time()
        if not ar.result.authenticated:
            msg = "Failed to authenticate with external login."
            self.log.error(f'{self.me}: {msg} in {(end_time-start_time)/1_000:.3f}s.')
            raise AuthenticationFailedException(msg)
        self.__set_cookie(ar)
        self.log.info(f'{self.me} authenticated with external login in {(end_time-start_time)/1_000:.3f}s.')

    def password(self, user: str, password: str, effective_user: str = None) -> None:
        """
        Authenticates to the server using a username and password.

        :param user: the user to authenticate
        :param password: the user's password
        :param effective_user: the user to operate as, defaults to the user to authenticate
        """
        uc = deephaven_enterprise.proto.auth_pb2.UserContext()
        uc.authenticatedUser = user
        if effective_user is None:
            uc.effectiveUser = user
        else:
            uc.effectiveUser = effective_user

        abp = deephaven_enterprise.proto.auth_pb2.AuthenticateByPasswordRequest()
        abp.client_id.CopyFrom(self.clientID)
        abp.user_context.CopyFrom(uc)
        abp.password = password
        start_time = time.time()
        ar = self.stub.authenticateByPassword(abp,
                                              timeout=self.rpc_timeout_secs, wait_for_ready=True)
        end_time = time.time()
        if not ar.result.authenticated:
            msg = "Failed to authenticate with password."
            self.log.error(f'{self.me}: {msg} in {(end_time-start_time)/1_000:.3f}s.')
            raise AuthenticationFailedException(msg)
        self.__set_cookie(ar)
        self.log.info(f'{self.me} authenticated with password in {(end_time-start_time)/1_000:.3f}s.')

    def saml(self, login_uri: str) -> None:
        """
        Authenticate using SAML, which must be configured on the server.

        :param login_uri: the URI for the Deephaven SAML plugin.  Often "https://deephaven-auth-server:9032/dh-saml/", but is dependent on configuration.
        """
        random_bytes = secrets.token_bytes(96)
        nonce = base64.b64encode(random_bytes)
        uri = login_uri + "?key=" + urllib.parse.quote_plus(nonce)
        webbrowser.open(uri)

        try:
            self.external_login(nonce)
        except AuthenticationFailedException as e:
            msg = "Failed to authenticate with SAML."
            raise AuthenticationFailedException(msg) from e

    @staticmethod
    def generate_keypair(user: str) -> (str, str):
        """
        Generate a new keypair in Deephaven format.

        The private key can be stored as a file and used to authenticate using the private_key method.  The public key
        contains a username followed by the public key.  The public key should be uploaded to the Deephaven ACL write
        server.

        :param user: the username to write into the key
        :return: a tuple containing the private key text and the public key text in Deephaven format
        """
        key = Cryptodome.PublicKey.ECC.generate(curve='p256')
        ec_sentinel = "EC:".encode("utf-8")
        pub = base64.b64encode(ec_sentinel + key.public_key().export_key(
            format="DER")).decode("us-ascii")
        priv = base64.b64encode(ec_sentinel + key.export_key(
            format="DER", use_pkcs8=True)).decode("us-ascii")
        privtext = """user %s
operateas %s
public %s
private %s
""" % (user, user, pub, priv)
        pubtext = "%s %s" % (user, pub)
        return (privtext, pubtext)

    def upload_key(self, pubtext: str, url: str, delete: bool = False):
        """
        Upload a public key to the ACL write server.

        :param pubtext: text of the public key, in Deephaven format
        :param url: URL of the ACL write server (e.g., https://foo.bar.company.com:9044/acl/)
        :param delete: True to delete the key instead of add the key
        """
        lines: List[str] = list(filter(lambda y: len(y) > 0, map(
            lambda x: x.strip(), pubtext.splitlines())))
        if (len(lines) != 1):
            raise Exception(
                "Invalid public key text, must be one line of the form '<username> <base64 encoded key>'")

        auth_token = self.get_token("DbAclWriteServer")
        token_as_string = auth_token.SerializeToString()
        token_encoded: bytes = base64.b64encode(token_as_string)
        authorization_header = token_encoded.decode("us-ascii")
        (user, key) = pubtext.split(" ")

        headers = {"Authorization": authorization_header,
                   "Content-Type": 'application/json'}

        ec_sentinel = "EC:".encode("utf-8")
        pub_key_bytes = base64.b64decode(key)
        if pub_key_bytes[0:3] == ec_sentinel:
            algorithm = "EC"
        else:
            algorithm = "DSA"

        if delete:
            path = url + ("/" if not url.endswith("/") else "") + \
                "publickey/" + urllib.parse.quote_plus(user)
            query_params = {"encodedStr": key, "algorithm": algorithm}
            response: requests.Response = requests.delete(
                path, params=query_params, headers=headers)
            if response.status_code != 204:
                raise Exception("Unexpected response: " + str(response))
        else:
            entity = dict()
            entity["user"] = user
            entity["encodedStr"] = key
            entity["algorithm"] = algorithm

            path = url + ("/" if not url.endswith("/") else "") + "publickey/"

            response: requests.Response = requests.post(
                path, json=entity, headers=headers)
            if response.status_code != 201:
                raise Exception("Unexpected response: " + str(response))

    def private_key(self, file: Union[str, io.StringIO]) -> None:
        """
        Authenticate to the server using a Deephaven format private key file.

        https://deephaven.io/enterprise/docs/resources/how-to/connect-from-java/#instructions-for-setting-up-private-keys

        :param file: a string file name containing the private key produced by generate-iris-keys, or alternatively an
        io.StringIO instance (which may be closed after it is read)
        """
        # Read the key information from the file

        desc: str = None

        if isinstance(file, str):
            # 'file' is a string: assume it's the path to the private key file
            desc = file
            with open(file, 'r') as fp:
                keydict = self.__read_key_dict(fp)
        else:
            desc = "<StringIO>"
            with file:
                keydict = self.__read_key_dict(file)

        for field in ["public", "private", "user", "operateas"]:
            if field not in keydict:
                raise Exception(
                    "'%s' is not a Deephaven private key file, %s is not present" % (desc, field))

        try:
            # pkcs8 encoded keyspec
            public_pkcs8_bytes = base64.b64decode(keydict["public"])
            private_pkcs8_bytes = base64.b64decode(keydict["private"])

            ec_sentinel = "EC:".encode("utf-8")
            if private_pkcs8_bytes[0:3] == ec_sentinel:
                private_key = Cryptodome.PublicKey.ECC.import_key(
                    private_pkcs8_bytes[3:])
                public_key = Cryptodome.PublicKey.ECC.import_key(
                    public_pkcs8_bytes[3:])
            else:
                private_key = Cryptodome.PublicKey.DSA.import_key(
                    private_pkcs8_bytes)
                public_key = Cryptodome.PublicKey.DSA.import_key(
                    public_pkcs8_bytes)
        except Exception as e:
            msg = "Invalid private key file '%s'" % desc
            self.log.error(f'{self.me}: {msg}')
            raise Exception(msg) from e

        # Get the nonce
        nr = deephaven_enterprise.proto.auth_pb2.GetNonceRequest()
        nr.client_id.CopyFrom(self.clientID)
        gnr = self.stub.getNonce(nr)

        # Sign the nonce
        try:
            hash_of_nonce = Cryptodome.Hash.SHA256.new(gnr.nonce)
            signer = Cryptodome.Signature.DSS.new(
                private_key, 'fips-186-3', encoding="der")
            signature = signer.sign(hash_of_nonce)
        except Exception as e:
            msg = "Could not sign nonce with private key '%s'" % desc
            self.log.error(f'{self.me}: {msg}')
            raise Exception(msg) from e

        # Now verify our signature locally, just in case
        try:
            verifier = Cryptodome.Signature.DSS.new(
                public_key, 'fips-186-3', encoding="der")
            verifier.verify(hash_of_nonce, signature)
        except Exception as e:
            msg = "Could not verify our own signature with private key '%s'" % desc
            self.log.error(f'{self.me}: {msg}')
            raise Exception(msg) from e

        # Produce the authentication message
        abpk = deephaven_enterprise.proto.auth_pb2.AuthenticateByPublicKeyRequest()
        abpk.user_context.authenticatedUser = keydict["user"]
        abpk.user_context.effectiveUser = keydict["operateas"]
        abpk.client_id.CopyFrom(self.clientID)
        abpk.public_key = public_pkcs8_bytes
        abpk.challenge_response = signature
        abpk.ip_address = gnr.ip_address

        # Do the authentication
        start_time = time.time()
        response = self.stub.authenticateByPublicKey(abpk,
                                                     timeout=self.rpc_timeout_secs, wait_for_ready=True)
        end_time = time.time()
        if not response.result.authenticated:
            msg = "Failed to authenticate with private key."
            self.log.error(f'{self.me}: {msg} in {(end_time-start_time)/1_000:.3f}s.')
            raise AuthenticationFailedException(msg)
        self.__set_cookie(response)
        self.log.info(f'{self.me} authenticated with private key in {(end_time-start_time)/1_000:.3f}s.')

    def __read_key_dict(self, fp):
        keydict = {}
        for x in fp.readlines():
            stripped = x.split("#", 1)[0].strip()
            if len(stripped) == 0:
                continue
            kv = re.split("\\s+", stripped, 2)
            keydict[kv[0]] = kv[1]
        return keydict

    def get_token(self, service: str, timeout: float = None) -> deephaven_enterprise.proto.auth_pb2.Token:
        """
        Get an authentication token to present to another Deephaven service.  This token may only be used one time,
        as it is consumed by the authentication server during the verification process.

        :param service: the service that will verify the token (e.g., "PersistentQueryController")
        :param timeout: how long to wait for the gRPC call to complete
        :return: the token
        """
        gtr: deephaven_enterprise.proto.auth_pb2.GetTokenRequest = deephaven_enterprise.proto.auth_pb2.GetTokenRequest()
        gtr.service = service
        gtr.cookie = self.cookie
        return self.stub.getToken(gtr,
                                  timeout=self.rpc_timeout_secs, wait_for_ready=True).token

    def close(self) -> None:
        """
        Logout from the authentication server.  No further tokens may be requested by this client.
        """
        if self.cookie is not None:
            ivcr = deephaven_enterprise.proto.auth_pb2.InvalidateCookieRequest()
            ivcr.cookie = self.cookie
            self.cookie = None
            self.cookie_deadline_time_millis = None
            self.stub.invalidateCookie(ivcr,
                                       timeout=self.rpc_timeout_secs, wait_for_ready=True)
        if self.opened_channel:
            self.channel.close()

    def ping(self):
        """
        Pings the server, refreshing our cookie.
        :returns: True if a ping was sent, False if there is no active cookie.
        """
        cookie = self.cookie
        if cookie is None:
            return False
        rcreq = deephaven_enterprise.proto.auth_pb2.RefreshCookieRequest()
        rcreq.cookie = cookie
        rcresp = self.stub.refreshCookie(rcreq,
                                         timeout=self.rpc_timeout_secs, wait_for_ready=True)
        self.cookie_deadline_time_millis = rcresp.cookie_deadline_time_millis
        return True

    def _delegate_token(self, abdt: deephaven_enterprise.proto.auth_pb2.AuthenticateByDelegateTokenRequest):
        """
        Authenticate using the provided AuthenticateByDelegateTokenRequest

        Note: This method is for internal use only.

        :param abdt: an AuthenticateByDelegateTokenRequest that is sent to our gRPC stub
        """
        start_time = time.time()
        ar = self.stub.authenticateByDelegateToken(abdt)
        end_time = time.time()
        if not ar.result.authenticated:
            msg = "Failed to authenticate with delegate token."
            self.log.error(f'{self.me}: {msg} in {(end_time-start_time)/1_000:.3f}s.')
            raise deephaven_enterprise.client.auth.AuthenticationFailedException(msg)
        self.__set_cookie(ar)
        self.log.info(f'{self.me}: authenticated with delegate token in {(end_time-start_time)/1_000:.3f}s.')

    def __set_cookie(self, ar):
        self.cookie = ar.result.cookie
        self.cookie_deadline_time_millis = ar.result.cookie_deadline_time_millis
        self.__start_refresh()

    def __start_refresh(self):
        self.refresh_thread = RefreshThread(self)
        self.refresh_thread.start()


class AuthenticationFailedException(Exception):
    """
    This Exception is raised when the server responds to our authentication request with a failure (e.g. bad password or
    bad key).  Other errors, like the server not responding at all are not covered by this Exception.
    """
    pass



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: deephaven_enterprise/proto/auth.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%deephaven_enterprise/proto/auth.proto\x12\x17io.deephaven.proto.auth\"?\n\x0bUserContext\x12\x19\n\x11\x61uthenticatedUser\x18\x01 \x01(\t\x12\x15\n\reffectiveUser\x18\x02 \x01(\t\"&\n\x08\x43lientId\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04uuid\x18\x02 \x01(\x0c\"\x9d\x01\n\x14\x41uthenticationResult\x12\x15\n\rauthenticated\x18\x01 \x01(\x08\x12\x0e\n\x06\x63ookie\x18\x02 \x01(\x0c\x12#\n\x1b\x63ookie_deadline_time_millis\x18\x03 \x01(\x10\x12\x39\n\x0buserContext\x18\x04 \x01(\x0b\x32$.io.deephaven.proto.auth.UserContext\"z\n\x05Token\x12\x10\n\x08token_id\x18\x01 \x01(\x06\x12\x0f\n\x07service\x18\x02 \x01(\t\x12:\n\x0cuser_context\x18\x03 \x01(\x0b\x32$.io.deephaven.proto.auth.UserContext\x12\x12\n\nip_address\x18\x04 \x01(\x0c\"\x92\x01\n\"AuthenticateByDelegateTokenRequest\x12\x34\n\tclient_id\x18\x01 \x01(\x0b\x32!.io.deephaven.proto.auth.ClientId\x12\x36\n\x0e\x64\x65legate_token\x18\x02 \x01(\x0b\x32\x1e.io.deephaven.proto.auth.Token\"d\n#AuthenticateByDelegateTokenResponse\x12=\n\x06result\x18\x01 \x01(\x0b\x32-.io.deephaven.proto.auth.AuthenticationResult\"&\n\x14RefreshCookieRequest\x12\x0e\n\x06\x63ookie\x18\x01 \x01(\x0c\"P\n\x15RefreshCookieResponse\x12\x12\n\x06\x63ookie\x18\x01 \x01(\x0c\x42\x02\x18\x01\x12#\n\x1b\x63ookie_deadline_time_millis\x18\x02 \x01(\x10\")\n\x17InvalidateCookieRequest\x12\x0e\n\x06\x63ookie\x18\x01 \x01(\x0c\"\x1a\n\x18InvalidateCookieResponse\"\xa3\x01\n\x1d\x41uthenticateByPasswordRequest\x12\x34\n\tclient_id\x18\x01 \x01(\x0b\x32!.io.deephaven.proto.auth.ClientId\x12:\n\x0cuser_context\x18\x02 \x01(\x0b\x32$.io.deephaven.proto.auth.UserContext\x12\x10\n\x08password\x18\x03 \x01(\t\"_\n\x1e\x41uthenticateByPasswordResponse\x12=\n\x06result\x18\x01 \x01(\x0b\x32-.io.deephaven.proto.auth.AuthenticationResult\"\xd6\x01\n\x1e\x41uthenticateByPublicKeyRequest\x12\x34\n\tclient_id\x18\x01 \x01(\x0b\x32!.io.deephaven.proto.auth.ClientId\x12:\n\x0cuser_context\x18\x02 \x01(\x0b\x32$.io.deephaven.proto.auth.UserContext\x12\x12\n\npublic_key\x18\x03 \x01(\x0c\x12\x1a\n\x12\x63hallenge_response\x18\x04 \x01(\x0c\x12\x12\n\nip_address\x18\x05 \x01(\x0c\"`\n\x1f\x41uthenticateByPublicKeyResponse\x12=\n\x06result\x18\x01 \x01(\x0b\x32-.io.deephaven.proto.auth.AuthenticationResult\"b\n\x1d\x41uthenticateByExternalRequest\x12\x34\n\tclient_id\x18\x01 \x01(\x0b\x32!.io.deephaven.proto.auth.ClientId\x12\x0b\n\x03key\x18\x02 \x01(\t\"_\n\x1e\x41uthenticateByExternalResponse\x12=\n\x06result\x18\x01 \x01(\x0b\x32-.io.deephaven.proto.auth.AuthenticationResult\"\x9f\x01\n\x1b\x41uthenticateByCookieRequest\x12\x34\n\tclient_id\x18\x01 \x01(\x0b\x32!.io.deephaven.proto.auth.ClientId\x12:\n\x0cuser_context\x18\x02 \x01(\x0b\x32$.io.deephaven.proto.auth.UserContext\x12\x0e\n\x06\x63ookie\x18\x03 \x01(\x0c\"]\n\x1c\x41uthenticateByCookieResponse\x12=\n\x06result\x18\x01 \x01(\x0b\x32-.io.deephaven.proto.auth.AuthenticationResult\"G\n\x0fGetNonceRequest\x12\x34\n\tclient_id\x18\x01 \x01(\x0b\x32!.io.deephaven.proto.auth.ClientId\"5\n\x10GetNonceResponse\x12\r\n\x05nonce\x18\x01 \x01(\x0c\x12\x12\n\nip_address\x18\x02 \x01(\x0c\"2\n\x0fGetTokenRequest\x12\x0e\n\x06\x63ookie\x18\x01 \x01(\x0c\x12\x0f\n\x07service\x18\x02 \x01(\t\"A\n\x10GetTokenResponse\x12-\n\x05token\x18\x01 \x01(\x0b\x32\x1e.io.deephaven.proto.auth.Token\"K\n\x11GetTokenAsRequest\x12\x0e\n\x06\x63ookie\x18\x01 \x01(\x0c\x12\x0f\n\x07service\x18\x02 \x01(\t\x12\x15\n\roperateAsUser\x18\x03 \x01(\t\"C\n\x12GetTokenAsResponse\x12-\n\x05token\x18\x01 \x01(\x0b\x32\x1e.io.deephaven.proto.auth.Token\"T\n\x12VerifyTokenRequest\x12\x0f\n\x07service\x18\x01 \x01(\t\x12-\n\x05token\x18\x02 \x01(\x0b\x32\x1e.io.deephaven.proto.auth.Token\"*\n\x13VerifyTokenResponse\x12\x13\n\x0btoken_valid\x18\x01 \x01(\x08\"\x0f\n\rReloadRequest\" \n\x0eReloadResponse\x12\x0e\n\x06result\x18\x01 \x01(\t\"_\n\x0cKvClientData\x12\x14\n\x0chashedCookie\x18\x01 \x01(\x0c\x12\x39\n\x0buserContext\x18\x02 \x01(\x0b\x32$.io.deephaven.proto.auth.UserContext\"U\n\tAuthError\x12\x37\n\ncallStatus\x18\x01 \x01(\x0e\x32#.io.deephaven.proto.auth.CallStatus\x12\x0f\n\x07message\x18\x02 \x01(\t\"\xd5\x01\n\x16VerifyChallengeRequest\x12\x19\n\x11server_ip_address\x18\x01 \x01(\x0c\x12\x34\n\tclient_id\x18\x02 \x01(\x0b\x32!.io.deephaven.proto.auth.ClientId\x12:\n\x0cuser_context\x18\x03 \x01(\x0b\x32$.io.deephaven.proto.auth.UserContext\x12\x12\n\npublic_key\x18\x04 \x01(\x0c\x12\x1a\n\x12\x63hallenge_response\x18\x05 \x01(\x0c\"e\n\x17VerifyChallengeResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x39\n\x0buserContext\x18\x02 \x01(\x0b\x32$.io.deephaven.proto.auth.UserContext*\xcd\x01\n\nCallStatus\x12\x06\n\x02OK\x10\x00\x12\x1c\n\x18\x41UTH_ERROR_NON_RETRYABLE\x10\x01\x12\x18\n\x14\x41UTH_ERROR_RETRYABLE\x10\x02\x12\x19\n\x15\x41LREADY_AUTHENTICATED\x10\x03\x12\x15\n\x11NOT_AUTHENTICATED\x10\x04\x12\x10\n\x0cPUB_PRIV_KEY\x10\x05\x12\x12\n\x0eTOKEN_CREATION\x10\x06\x12\x16\n\x12TOKEN_VERIFICATION\x10\x07\x12\x0f\n\x0bUNAVAILABLE\x10\x08\x42\x04H\x01P\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'deephaven_enterprise.proto.auth_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'H\001P\001'
  _globals['_REFRESHCOOKIERESPONSE'].fields_by_name['cookie']._loaded_options = None
  _globals['_REFRESHCOOKIERESPONSE'].fields_by_name['cookie']._serialized_options = b'\030\001'
  _globals['_CALLSTATUS']._serialized_start=3009
  _globals['_CALLSTATUS']._serialized_end=3214
  _globals['_USERCONTEXT']._serialized_start=66
  _globals['_USERCONTEXT']._serialized_end=129
  _globals['_CLIENTID']._serialized_start=131
  _globals['_CLIENTID']._serialized_end=169
  _globals['_AUTHENTICATIONRESULT']._serialized_start=172
  _globals['_AUTHENTICATIONRESULT']._serialized_end=329
  _globals['_TOKEN']._serialized_start=331
  _globals['_TOKEN']._serialized_end=453
  _globals['_AUTHENTICATEBYDELEGATETOKENREQUEST']._serialized_start=456
  _globals['_AUTHENTICATEBYDELEGATETOKENREQUEST']._serialized_end=602
  _globals['_AUTHENTICATEBYDELEGATETOKENRESPONSE']._serialized_start=604
  _globals['_AUTHENTICATEBYDELEGATETOKENRESPONSE']._serialized_end=704
  _globals['_REFRESHCOOKIEREQUEST']._serialized_start=706
  _globals['_REFRESHCOOKIEREQUEST']._serialized_end=744
  _globals['_REFRESHCOOKIERESPONSE']._serialized_start=746
  _globals['_REFRESHCOOKIERESPONSE']._serialized_end=826
  _globals['_INVALIDATECOOKIEREQUEST']._serialized_start=828
  _globals['_INVALIDATECOOKIEREQUEST']._serialized_end=869
  _globals['_INVALIDATECOOKIERESPONSE']._serialized_start=871
  _globals['_INVALIDATECOOKIERESPONSE']._serialized_end=897
  _globals['_AUTHENTICATEBYPASSWORDREQUEST']._serialized_start=900
  _globals['_AUTHENTICATEBYPASSWORDREQUEST']._serialized_end=1063
  _globals['_AUTHENTICATEBYPASSWORDRESPONSE']._serialized_start=1065
  _globals['_AUTHENTICATEBYPASSWORDRESPONSE']._serialized_end=1160
  _globals['_AUTHENTICATEBYPUBLICKEYREQUEST']._serialized_start=1163
  _globals['_AUTHENTICATEBYPUBLICKEYREQUEST']._serialized_end=1377
  _globals['_AUTHENTICATEBYPUBLICKEYRESPONSE']._serialized_start=1379
  _globals['_AUTHENTICATEBYPUBLICKEYRESPONSE']._serialized_end=1475
  _globals['_AUTHENTICATEBYEXTERNALREQUEST']._serialized_start=1477
  _globals['_AUTHENTICATEBYEXTERNALREQUEST']._serialized_end=1575
  _globals['_AUTHENTICATEBYEXTERNALRESPONSE']._serialized_start=1577
  _globals['_AUTHENTICATEBYEXTERNALRESPONSE']._serialized_end=1672
  _globals['_AUTHENTICATEBYCOOKIEREQUEST']._serialized_start=1675
  _globals['_AUTHENTICATEBYCOOKIEREQUEST']._serialized_end=1834
  _globals['_AUTHENTICATEBYCOOKIERESPONSE']._serialized_start=1836
  _globals['_AUTHENTICATEBYCOOKIERESPONSE']._serialized_end=1929
  _globals['_GETNONCEREQUEST']._serialized_start=1931
  _globals['_GETNONCEREQUEST']._serialized_end=2002
  _globals['_GETNONCERESPONSE']._serialized_start=2004
  _globals['_GETNONCERESPONSE']._serialized_end=2057
  _globals['_GETTOKENREQUEST']._serialized_start=2059
  _globals['_GETTOKENREQUEST']._serialized_end=2109
  _globals['_GETTOKENRESPONSE']._serialized_start=2111
  _globals['_GETTOKENRESPONSE']._serialized_end=2176
  _globals['_GETTOKENASREQUEST']._serialized_start=2178
  _globals['_GETTOKENASREQUEST']._serialized_end=2253
  _globals['_GETTOKENASRESPONSE']._serialized_start=2255
  _globals['_GETTOKENASRESPONSE']._serialized_end=2322
  _globals['_VERIFYTOKENREQUEST']._serialized_start=2324
  _globals['_VERIFYTOKENREQUEST']._serialized_end=2408
  _globals['_VERIFYTOKENRESPONSE']._serialized_start=2410
  _globals['_VERIFYTOKENRESPONSE']._serialized_end=2452
  _globals['_RELOADREQUEST']._serialized_start=2454
  _globals['_RELOADREQUEST']._serialized_end=2469
  _globals['_RELOADRESPONSE']._serialized_start=2471
  _globals['_RELOADRESPONSE']._serialized_end=2503
  _globals['_KVCLIENTDATA']._serialized_start=2505
  _globals['_KVCLIENTDATA']._serialized_end=2600
  _globals['_AUTHERROR']._serialized_start=2602
  _globals['_AUTHERROR']._serialized_end=2687
  _globals['_VERIFYCHALLENGEREQUEST']._serialized_start=2690
  _globals['_VERIFYCHALLENGEREQUEST']._serialized_end=2903
  _globals['_VERIFYCHALLENGERESPONSE']._serialized_start=2905
  _globals['_VERIFYCHALLENGERESPONSE']._serialized_end=3006
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: deephaven_enterprise/proto/persistent_query.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from deephaven_enterprise.proto import common_pb2 as deephaven__enterprise_dot_proto_dot_common__pb2
from deephaven_enterprise.proto import controller_common_pb2 as deephaven__enterprise_dot_proto_dot_controller__common__pb2
from deephaven_enterprise.proto import table_definition_pb2 as deephaven__enterprise_dot_proto_dot_table__definition__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n1deephaven_enterprise/proto/persistent_query.proto\x12\x1dio.deephaven.proto.controller\x1a\'deephaven_enterprise/proto/common.proto\x1a\x32\x64\x65\x65phaven_enterprise/proto/controller_common.proto\x1a\x31\x64\x65\x65phaven_enterprise/proto/table_definition.proto\"\xc6\x08\n\x1cPersistentQueryConfigMessage\x12\x0e\n\x06serial\x18\x01 \x01(\x03\x12\x0f\n\x07version\x18\x02 \x01(\x03\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\r\n\x05owner\x18\x04 \x01(\t\x12\x0f\n\x07\x65nabled\x18\x05 \x01(\x08\x12\x12\n\nheapSizeGb\x18\x06 \x01(\x01\x12\x1d\n\x15\x62ufferPoolToHeapRatio\x18\x07 \x01(\x01\x12 \n\x18\x64\x65tailedGCLoggingEnabled\x18\x08 \x01(\x08\x12\x19\n\x11\x65xtraJvmArguments\x18\t \x03(\t\x12!\n\x19\x65xtraEnvironmentVariables\x18\n \x03(\t\x12\x1a\n\x12\x63lassPathAdditions\x18\x0b \x03(\t\x12\x12\n\nserverName\x18\x0c \x01(\t\x12\x13\n\x0b\x61\x64minGroups\x18\r \x03(\t\x12\x14\n\x0cviewerGroups\x18\x0e \x03(\t\x12\x45\n\x0crestartUsers\x18\x0f \x01(\x0e\x32/.io.deephaven.proto.controller.RestartUsersEnum\x12\x14\n\nscriptCode\x18\x10 \x01(\tH\x00\x12\x14\n\nscriptPath\x18\x11 \x01(\tH\x00\x12\x16\n\x0escriptLanguage\x18\x12 \x01(\t\x12\x19\n\x11\x63onfigurationType\x18\x13 \x01(\t\x12#\n\x16typeSpecificFieldsJson\x18\x14 \x01(\tH\x01\x88\x01\x01\x12\x12\n\nscheduling\x18\x15 \x03(\t\x12\x14\n\x0ctimeoutNanos\x18\x16 \x01(\x03\x12\x12\n\njvmProfile\x18\x17 \x01(\t\x12#\n\x1blastModifiedByAuthenticated\x18\x18 \x01(\t\x12\x1f\n\x17lastModifiedByEffective\x18\x19 \x01(\t\x12\x1d\n\x15lastModifiedTimeNanos\x18\x1a \x01(\x03\x12\x17\n\x0f\x63ompletedStatus\x18\x1b \x01(\t\x12\x1b\n\x13\x65xpirationTimeNanos\x18\x1c \x01(\x03\x12\x19\n\x11kubernetesControl\x18\x1d \x01(\t\x12\x12\n\nworkerKind\x18\x1e \x01(\t\x12\x18\n\x10\x63reatedTimeNanos\x18\x1f \x01(\x03\x12\x14\n\x0creplicaCount\x18  \x01(\x05\x12\x12\n\nspareCount\x18! \x01(\x05\x12\x18\n\x10\x61ssignmentPolicy\x18\" \x01(\t\x12#\n\x16\x61ssignmentPolicyParams\x18# \x01(\tH\x02\x88\x01\x01\x12\x1a\n\x12\x61\x64\x64itionalMemoryGb\x18$ \x01(\x01\x12\x15\n\rpythonControl\x18% \x01(\t\x12\x1c\n\x14genericWorkerControl\x18& \x01(\tB\x0c\n\nscriptDataB\x19\n\x17_typeSpecificFieldsJsonB\x19\n\x17_assignmentPolicyParams\"\xd0\x01\n\x19\x45xportedObjectInfoMessage\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x43\n\x04type\x18\x02 \x01(\x0e\x32\x35.io.deephaven.proto.controller.ExportedObjectTypeEnum\x12J\n\x0ftableDefinition\x18\x03 \x01(\x0b\x32\x31.io.deephaven.proto.common.TableDefinitionMessage\x12\x14\n\x0coriginalType\x18\x04 \x01(\t\"3\n\x15WorkerProtocolMessage\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x05\"\x87\x02\n!ProcessorConnectionDetailsMessage\x12G\n\tprotocols\x18\x01 \x03(\x0b\x32\x34.io.deephaven.proto.controller.WorkerProtocolMessage\x12\x12\n\nworkerName\x18\x02 \x01(\t\x12\x15\n\rprocessInfoId\x18\x03 \x01(\t\x12\x15\n\rprocessorHost\x18\x04 \x01(\t\x12\x13\n\x0b\x65nvoyPrefix\x18\x05 \x01(\t\x12\x0f\n\x07grpcUrl\x18\x06 \x01(\t\x12\x11\n\tstaticUrl\x18\x07 \x01(\t\x12\x1e\n\x16\x65nterpriseWebSocketUrl\x18\x08 \x01(\t\"\xf6\x07\n\x1bPersistentQueryStateMessage\x12\x0e\n\x06serial\x18\x01 \x01(\x03\x12\x0f\n\x07version\x18\x02 \x01(\x03\x12H\n\x06status\x18\x03 \x01(\x0e\x32\x38.io.deephaven.proto.controller.PersistentQueryStatusEnum\x12 \n\x18initializationStartNanos\x18\x04 \x01(\x03\x12#\n\x1binitializationCompleteNanos\x18\x05 \x01(\x03\x12\x17\n\x0flastUpdateNanos\x18\x06 \x01(\x03\x12\x16\n\x0e\x64ispatcherHost\x18\x07 \x01(\t\x12\x43\n\x0btableGroups\x18\x08 \x03(\x0b\x32..io.deephaven.proto.controller.NamedStringList\x12L\n\nscopeTypes\x18\t \x03(\x0b\x32\x38.io.deephaven.proto.controller.ExportedObjectInfoMessage\x12`\n\x11\x63onnectionDetails\x18\n \x01(\x0b\x32@.io.deephaven.proto.controller.ProcessorConnectionDetailsMessageH\x00\x88\x01\x01\x12Q\n\x10\x65xceptionDetails\x18\x0b \x01(\x0b\x32\x32.io.deephaven.proto.common.ExceptionDetailsMessageH\x01\x88\x01\x01\x12\x1d\n\x15typeSpecificStateJson\x18\x0c \x01(\t\x12\x1d\n\x15lastAuthenticatedUser\x18\r \x01(\t\x12\x19\n\x11lastEffectiveUser\x18\x0e \x01(\t\x12\"\n\x15scriptLoaderStateJson\x18\x0f \x01(\tH\x02\x88\x01\x01\x12\x13\n\x0bhasProgress\x18\x10 \x01(\x08\x12\x15\n\rprogressValue\x18\x11 \x01(\x05\x12\x17\n\x0fprogressMessage\x18\x12 \x01(\t\x12\x15\n\rengineVersion\x18\x13 \x01(\t\x12\x16\n\x0e\x64ispatcherPort\x18\x14 \x01(\x05\x12\x17\n\x0fshouldStopNanos\x18\x15 \x01(\x03\x12\x13\n\x0bnumFailures\x18\x16 \x01(\x05\x12\x1c\n\x14lastFailureTimeNanos\x18\x17 \x01(\x03\x12\x13\n\x0breplicaSlot\x18\x18 \x01(\x05\x12\x15\n\rstatusDetails\x18\x19 \x01(\tB\x14\n\x12_connectionDetailsB\x13\n\x11_exceptionDetailsB\x18\n\x16_scriptLoaderStateJson\"\xe9\x02\n\x1aPersistentQueryInfoMessage\x12K\n\x06\x63onfig\x18\x01 \x01(\x0b\x32;.io.deephaven.proto.controller.PersistentQueryConfigMessage\x12N\n\x05state\x18\x02 \x01(\x0b\x32:.io.deephaven.proto.controller.PersistentQueryStateMessageH\x00\x88\x01\x01\x12L\n\x08replicas\x18\x05 \x03(\x0b\x32:.io.deephaven.proto.controller.PersistentQueryStateMessage\x12J\n\x06spares\x18\x06 \x03(\x0b\x32:.io.deephaven.proto.controller.PersistentQueryStateMessageB\x08\n\x06_stateJ\x04\x08\x03\x10\x04J\x04\x08\x04\x10\x05*{\n\x10RestartUsersEnum\x12\x12\n\x0eRU_UNSPECIFIED\x10\x00\x12\x0c\n\x08RU_ADMIN\x10\x01\x12\x18\n\x14RU_ADMIN_AND_VIEWERS\x10\x02\x12\x11\n\rRU_RESERVED_1\x10\x03\x12\x18\n\x14RU_VIEWERS_WHEN_DOWN\x10\x04*\xb2\x02\n\x19PersistentQueryStatusEnum\x12\x13\n\x0fPQS_UNSPECIFIED\x10\x00\x12\x15\n\x11PQS_UNINITIALIZED\x10\x01\x12\x12\n\x0ePQS_CONNECTING\x10\x02\x12\x16\n\x12PQS_AUTHENTICATING\x10\x03\x12\x18\n\x14PQS_ACQUIRING_WORKER\x10\x04\x12\x14\n\x10PQS_INITIALIZING\x10\x05\x12\x0f\n\x0bPQS_RUNNING\x10\x06\x12\x0e\n\nPQS_FAILED\x10\x07\x12\r\n\tPQS_ERROR\x10\x08\x12\x14\n\x10PQS_DISCONNECTED\x10\t\x12\x10\n\x0cPQS_STOPPING\x10\n\x12\x0f\n\x0bPQS_STOPPED\x10\x0b\x12\x11\n\rPQS_COMPLETED\x10\x0c\x12\x11\n\rPQS_EXECUTING\x10\r*\xcf\x01\n\x16\x45xportedObjectTypeEnum\x12\x13\n\x0f\x45OT_UNSPECIFIED\x10\x00\x12\r\n\tEOT_TABLE\x10\x01\x12\x12\n\x0e\x45OT_TREE_TABLE\x10\x02\x12\x11\n\rEOT_TABLE_MAP\x10\x03\x12\x0e\n\nEOT_FIGURE\x10\x04\x12\x14\n\x10\x45OT_OTHER_WIDGET\x10\x05\x12\x1b\n\x17\x45OT_PARAMETERIZED_QUERY\x10\x06\x12\x0e\n\nEOT_PANDAS\x10\x07\x12\x17\n\x13\x45OT_NON_DISPLAYABLE\x10\x08\x42\x04H\x01P\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'deephaven_enterprise.proto.persistent_query_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'H\001P\001'
  _globals['_RESTARTUSERSENUM']._serialized_start=3236
  _globals['_RESTARTUSERSENUM']._serialized_end=3359
  _globals['_PERSISTENTQUERYSTATUSENUM']._serialized_start=3362
  _globals['_PERSISTENTQUERYSTATUSENUM']._serialized_end=3668
  _globals['_EXPORTEDOBJECTTYPEENUM']._serialized_start=3671
  _globals['_EXPORTEDOBJECTTYPEENUM']._serialized_end=3878
  _globals['_PERSISTENTQUERYCONFIGMESSAGE']._serialized_start=229
  _globals['_PERSISTENTQUERYCONFIGMESSAGE']._serialized_end=1323
  _globals['_EXPORTEDOBJECTINFOMESSAGE']._serialized_start=1326
  _globals['_EXPORTEDOBJECTINFOMESSAGE']._serialized_end=1534
  _globals['_WORKERPROTOCOLMESSAGE']._serialized_start=1536
  _globals['_WORKERPROTOCOLMESSAGE']._serialized_end=1587
  _globals['_PROCESSORCONNECTIONDETAILSMESSAGE']._serialized_start=1590
  _globals['_PROCESSORCONNECTIONDETAILSMESSAGE']._serialized_end=1853
  _globals['_PERSISTENTQUERYSTATEMESSAGE']._serialized_start=1856
  _globals['_PERSISTENTQUERYSTATEMESSAGE']._serialized_end=2870
  _globals['_PERSISTENTQUERYINFOMESSAGE']._serialized_start=2873
  _globals['_PERSISTENTQUERYINFOMESSAGE']._serialized_end=3234
# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from deephaven_enterprise.proto import acl_pb2 as deephaven__enterprise_dot_proto_dot_acl__pb2
from deephaven_enterprise.proto import auth_pb2 as deephaven__enterprise_dot_proto_dot_auth__pb2
from deephaven_enterprise.proto import controller_common_pb2 as deephaven__enterprise_dot_proto_dot_controller__common__pb2
from deephaven_enterprise.proto import controller_pb2 as deephaven__enterprise_dot_proto_dot_controller__pb2

GRPC_GENERATED_VERSION = '1.63.0'
GRPC_VERSION = grpc.__version__
EXPECTED_ERROR_RELEASE = '1.65.0'
SCHEDULED_RELEASE_DATE = 'June 25, 2024'
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    warnings.warn(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in deephaven_enterprise/proto/controller_service_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
        + f' This warning will become an error in {EXPECTED_ERROR_RELEASE},'
        + f' scheduled for release on {SCHEDULED_RELEASE_DATE}.',
        RuntimeWarning
    )


class ControllerApiStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.authenticate = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.ControllerApi/authenticate',
                request_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.AuthenticationRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.AuthenticationResponse.FromString,
                _registered_method=True)
        self.addQuery = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.ControllerApi/addQuery',
                request_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.AddQueryRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.AddQueryResponse.FromString,
                _registered_method=True)
        self.removeQuery = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.ControllerApi/removeQuery',
                request_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.RemoveQueryRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.RemoveQueryResponse.FromString,
                _registered_method=True)
        self.modifyQuery = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.ControllerApi/modifyQuery',
                request_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.ModifyQueryRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.ModifyQueryResponse.FromString,
                _registered_method=True)
        self.restartQuery = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.ControllerApi/restartQuery',
                request_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.RestartQueryRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.RestartQueryResponse.FromString,
                _registered_method=True)
        self.stopQuery = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.ControllerApi/stopQuery',
                request_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.StopQueryRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.StopQueryResponse.FromString,
                _registered_method=True)
        self.restartReplica = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.ControllerApi/restartReplica',
                request_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.RestartReplicaRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.RestartReplicaResponse.FromString,
                _registered_method=True)
        self.stopReplica = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.ControllerApi/stopReplica',
                request_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.StopReplicaRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.StopReplicaResponse.FromString,
                _registered_method=True)
        self.determineDispatcher = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.ControllerApi/determineDispatcher',
                request_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.DetermineDispatcherRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.DetermineDispatcherResponse.FromString,
                _registered_method=True)
        self.selectionProviderStatus = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.ControllerApi/selectionProviderStatus',
                request_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.ServerSelectionStatusRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.ServerSelectionStatusResponse.FromString,
                _registered_method=True)
        self.selectionProviderAdmin = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.ControllerApi/selectionProviderAdmin',
                request_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.ServerSelectionAdminRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.ServerSelectionAdminResponse.FromString,
                _registered_method=True)
        self.getConfiguration = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.ControllerApi/getConfiguration',
                request_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.GetConfigurationRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.GetConfigurationResponse.FromString,
                _registered_method=True)
        self.getScriptPaths = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.ControllerApi/getScriptPaths',
                request_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.GetScriptPathsRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.GetScriptPathsResponse.FromString,
                _registered_method=True)
        self.getScript = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.ControllerApi/getScript',
                request_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.GetScriptRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.GetScriptResponse.FromString,
                _registered_method=True)
        self.statusUpdate = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.ControllerApi/statusUpdate',
                request_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.StatusUpdateRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.StatusUpdateResponse.FromString,
                _registered_method=True)
        self.reloadConfig = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.ControllerApi/reloadConfig',
                request_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.ReloadConfigRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.ReloadConfigResponse.FromString,
                _registered_method=True)
        self.ping = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.ControllerApi/ping',
                request_serializer=deephaven__enterprise_dot_proto_dot_controller__common__pb2.PingRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_controller__common__pb2.PingResponse.FromString,
                _registered_method=True)
        self.subscribe = channel.unary_stream(
                '/io.deephaven.proto.controller.grpc.ControllerApi/subscribe',
                request_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.SubscribeRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.SubscribeResponse.FromString,
                _registered_method=True)
        self.invalidateCookie = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.ControllerApi/invalidateCookie',
                request_serializer=deephaven__enterprise_dot_proto_dot_auth__pb2.InvalidateCookieRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_auth__pb2.InvalidateCookieResponse.FromString,
                _registered_method=True)


class ControllerApiServicer(object):
    """Missing associated documentation comment in .proto file."""

    def authenticate(self, request, context):
        """Authentication
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def addQuery(self, request, context):
        """These are for query management
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def removeQuery(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def modifyQuery(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def restartQuery(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def stopQuery(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def restartReplica(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def stopReplica(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def determineDispatcher(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def selectionProviderStatus(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def selectionProviderAdmin(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getConfiguration(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getScriptPaths(self, request, context):
        """Script commands
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getScript(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def statusUpdate(self, request, context):
        """Configuration commands
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def reloadConfig(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ping(self, request, context):
        """Heartbeating
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def subscribe(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def invalidateCookie(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ControllerApiServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'authenticate': grpc.unary_unary_rpc_method_handler(
                    servicer.authenticate,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.AuthenticationRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.AuthenticationResponse.SerializeToString,
            ),
            'addQuery': grpc.unary_unary_rpc_method_handler(
                    servicer.addQuery,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.AddQueryRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.AddQueryResponse.SerializeToString,
            ),
            'removeQuery': grpc.unary_unary_rpc_method_handler(
                    servicer.removeQuery,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.RemoveQueryRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.RemoveQueryResponse.SerializeToString,
            ),
            'modifyQuery': grpc.unary_unary_rpc_method_handler(
                    servicer.modifyQuery,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.ModifyQueryRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.ModifyQueryResponse.SerializeToString,
            ),
            'restartQuery': grpc.unary_unary_rpc_method_handler(
                    servicer.restartQuery,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.RestartQueryRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.RestartQueryResponse.SerializeToString,
            ),
            'stopQuery': grpc.unary_unary_rpc_method_handler(
                    servicer.stopQuery,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.StopQueryRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.StopQueryResponse.SerializeToString,
            ),
            'restartReplica': grpc.unary_unary_rpc_method_handler(
                    servicer.restartReplica,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.RestartReplicaRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.RestartReplicaResponse.SerializeToString,
            ),
            'stopReplica': grpc.unary_unary_rpc_method_handler(
                    servicer.stopReplica,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.StopReplicaRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.StopReplicaResponse.SerializeToString,
            ),
            'determineDispatcher': grpc.unary_unary_rpc_method_handler(
                    servicer.determineDispatcher,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.DetermineDispatcherRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.DetermineDispatcherResponse.SerializeToString,
            ),
            'selectionProviderStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.selectionProviderStatus,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.ServerSelectionStatusRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.ServerSelectionStatusResponse.SerializeToString,
            ),
            'selectionProviderAdmin': grpc.unary_unary_rpc_method_handler(
                    servicer.selectionProviderAdmin,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.ServerSelectionAdminRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.ServerSelectionAdminResponse.SerializeToString,
            ),
            'getConfiguration': grpc.unary_unary_rpc_method_handler(
                    servicer.getConfiguration,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.GetConfigurationRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.GetConfigurationResponse.SerializeToString,
            ),
            'getScriptPaths': grpc.unary_unary_rpc_method_handler(
                    servicer.getScriptPaths,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.GetScriptPathsRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.GetScriptPathsResponse.SerializeToString,
            ),
            'getScript': grpc.unary_unary_rpc_method_handler(
                    servicer.getScript,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.GetScriptRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.GetScriptResponse.SerializeToString,
            ),
            'statusUpdate': grpc.unary_unary_rpc_method_handler(
                    servicer.statusUpdate,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.StatusUpdateRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.StatusUpdateResponse.SerializeToString,
            ),
            'reloadConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.reloadConfig,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.ReloadConfigRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.ReloadConfigResponse.SerializeToString,
            ),
            'ping': grpc.unary_unary_rpc_method_handler(
                    servicer.ping,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_controller__common__pb2.PingRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_controller__common__pb2.PingResponse.SerializeToString,
            ),
            'subscribe': grpc.unary_stream_rpc_method_handler(
                    servicer.subscribe,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_controller__pb2.SubscribeRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_controller__pb2.SubscribeResponse.SerializeToString,
            ),
            'invalidateCookie': grpc.unary_unary_rpc_method_handler(
                    servicer.invalidateCookie,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_auth__pb2.InvalidateCookieRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_auth__pb2.InvalidateCookieResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'io.deephaven.proto.controller.grpc.ControllerApi', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ControllerApi(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def authenticate(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.ControllerApi/authenticate',
            deephaven__enterprise_dot_proto_dot_controller__pb2.AuthenticationRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_controller__pb2.AuthenticationResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def addQuery(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.ControllerApi/addQuery',
            deephaven__enterprise_dot_proto_dot_controller__pb2.AddQueryRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_controller__pb2.AddQueryResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def removeQuery(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.ControllerApi/removeQuery',
            deephaven__enterprise_dot_proto_dot_controller__pb2.RemoveQueryRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_controller__pb2.RemoveQueryResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def modifyQuery(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.ControllerApi/modifyQuery',
            deephaven__enterprise_dot_proto_dot_controller__pb2.ModifyQueryRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_controller__pb2.ModifyQueryResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def restartQuery(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.ControllerApi/restartQuery',
            deephaven__enterprise_dot_proto_dot_controller__pb2.RestartQueryRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_controller__pb2.RestartQueryResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def stopQuery(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.ControllerApi/stopQuery',
            deephaven__enterprise_dot_proto_dot_controller__pb2.StopQueryRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_controller__pb2.StopQueryResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def restartReplica(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.ControllerApi/restartReplica',
            deephaven__enterprise_dot_proto_dot_controller__pb2.RestartReplicaRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_controller__pb2.RestartReplicaResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def stopReplica(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.ControllerApi/stopReplica',
            deephaven__enterprise_dot_proto_dot_controller__pb2.StopReplicaRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_controller__pb2.StopReplicaResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def determineDispatcher(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.ControllerApi/determineDispatcher',
            deephaven__enterprise_dot_proto_dot_controller__pb2.DetermineDispatcherRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_controller__pb2.DetermineDispatcherResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def selectionProviderStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.ControllerApi/selectionProviderStatus',
            deephaven__enterprise_dot_proto_dot_controller__pb2.ServerSelectionStatusRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_controller__pb2.ServerSelectionStatusResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def selectionProviderAdmin(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.ControllerApi/selectionProviderAdmin',
            deephaven__enterprise_dot_proto_dot_controller__pb2.ServerSelectionAdminRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_controller__pb2.ServerSelectionAdminResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getConfiguration(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.ControllerApi/getConfiguration',
            deephaven__enterprise_dot_proto_dot_controller__pb2.GetConfigurationRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_controller__pb2.GetConfigurationResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getScriptPaths(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.ControllerApi/getScriptPaths',
            deephaven__enterprise_dot_proto_dot_controller__pb2.GetScriptPathsRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_controller__pb2.GetScriptPathsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getScript(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.ControllerApi/getScript',
            deephaven__enterprise_dot_proto_dot_controller__pb2.GetScriptRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_controller__pb2.GetScriptResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def statusUpdate(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.ControllerApi/statusUpdate',
            deephaven__enterprise_dot_proto_dot_controller__pb2.StatusUpdateRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_controller__pb2.StatusUpdateResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def reloadConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.ControllerApi/reloadConfig',
            deephaven__enterprise_dot_proto_dot_controller__pb2.ReloadConfigRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_controller__pb2.ReloadConfigResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ping(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.ControllerApi/ping',
            deephaven__enterprise_dot_proto_dot_controller__common__pb2.PingRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_controller__common__pb2.PingResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def subscribe(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.ControllerApi/subscribe',
            deephaven__enterprise_dot_proto_dot_controller__pb2.SubscribeRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_controller__pb2.SubscribeResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def invalidateCookie(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.ControllerApi/invalidateCookie',
            deephaven__enterprise_dot_proto_dot_auth__pb2.InvalidateCookieRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_auth__pb2.InvalidateCookieResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class AclApiStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.isSuperUser = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.AclApi/isSuperUser',
                request_serializer=deephaven__enterprise_dot_proto_dot_acl__pb2.UnaryUserRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_acl__pb2.BooleanResponse.FromString,
                _registered_method=True)
        self.isAclEditor = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.AclApi/isAclEditor',
                request_serializer=deephaven__enterprise_dot_proto_dot_acl__pb2.UnaryUserRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_acl__pb2.BooleanResponse.FromString,
                _registered_method=True)
        self.isQueryManager = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.AclApi/isQueryManager',
                request_serializer=deephaven__enterprise_dot_proto_dot_acl__pb2.UnaryUserRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_acl__pb2.BooleanResponse.FromString,
                _registered_method=True)
        self.isSchemaManager = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.AclApi/isSchemaManager',
                request_serializer=deephaven__enterprise_dot_proto_dot_acl__pb2.UnaryUserRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_acl__pb2.BooleanResponse.FromString,
                _registered_method=True)
        self.getUsersForGroup = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.AclApi/getUsersForGroup',
                request_serializer=deephaven__enterprise_dot_proto_dot_acl__pb2.UnaryGroupRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_acl__pb2.UserOrGroupList.FromString,
                _registered_method=True)
        self.getGroupsForUser = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.AclApi/getGroupsForUser',
                request_serializer=deephaven__enterprise_dot_proto_dot_acl__pb2.UnaryUserRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_acl__pb2.UserOrGroupList.FromString,
                _registered_method=True)
        self.getAllUsers = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.AclApi/getAllUsers',
                request_serializer=deephaven__enterprise_dot_proto_dot_acl__pb2.GetAllUsersRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_acl__pb2.UserOrGroupList.FromString,
                _registered_method=True)
        self.getAllGroups = channel.unary_unary(
                '/io.deephaven.proto.controller.grpc.AclApi/getAllGroups',
                request_serializer=deephaven__enterprise_dot_proto_dot_acl__pb2.GetAllGroupsRequest.SerializeToString,
                response_deserializer=deephaven__enterprise_dot_proto_dot_acl__pb2.UserOrGroupList.FromString,
                _registered_method=True)


class AclApiServicer(object):
    """Missing associated documentation comment in .proto file."""

    def isSuperUser(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def isAclEditor(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def isQueryManager(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def isSchemaManager(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getUsersForGroup(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getGroupsForUser(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getAllUsers(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getAllGroups(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_AclApiServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'isSuperUser': grpc.unary_unary_rpc_method_handler(
                    servicer.isSuperUser,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_acl__pb2.UnaryUserRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_acl__pb2.BooleanResponse.SerializeToString,
            ),
            'isAclEditor': grpc.unary_unary_rpc_method_handler(
                    servicer.isAclEditor,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_acl__pb2.UnaryUserRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_acl__pb2.BooleanResponse.SerializeToString,
            ),
            'isQueryManager': grpc.unary_unary_rpc_method_handler(
                    servicer.isQueryManager,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_acl__pb2.UnaryUserRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_acl__pb2.BooleanResponse.SerializeToString,
            ),
            'isSchemaManager': grpc.unary_unary_rpc_method_handler(
                    servicer.isSchemaManager,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_acl__pb2.UnaryUserRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_acl__pb2.BooleanResponse.SerializeToString,
            ),
            'getUsersForGroup': grpc.unary_unary_rpc_method_handler(
                    servicer.getUsersForGroup,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_acl__pb2.UnaryGroupRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_acl__pb2.UserOrGroupList.SerializeToString,
            ),
            'getGroupsForUser': grpc.unary_unary_rpc_method_handler(
                    servicer.getGroupsForUser,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_acl__pb2.UnaryUserRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_acl__pb2.UserOrGroupList.SerializeToString,
            ),
            'getAllUsers': grpc.unary_unary_rpc_method_handler(
                    servicer.getAllUsers,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_acl__pb2.GetAllUsersRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_acl__pb2.UserOrGroupList.SerializeToString,
            ),
            'getAllGroups': grpc.unary_unary_rpc_method_handler(
                    servicer.getAllGroups,
                    request_deserializer=deephaven__enterprise_dot_proto_dot_acl__pb2.GetAllGroupsRequest.FromString,
                    response_serializer=deephaven__enterprise_dot_proto_dot_acl__pb2.UserOrGroupList.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'io.deephaven.proto.controller.grpc.AclApi', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class AclApi(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def isSuperUser(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.AclApi/isSuperUser',
            deephaven__enterprise_dot_proto_dot_acl__pb2.UnaryUserRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_acl__pb2.BooleanResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def isAclEditor(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.AclApi/isAclEditor',
            deephaven__enterprise_dot_proto_dot_acl__pb2.UnaryUserRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_acl__pb2.BooleanResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def isQueryManager(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.AclApi/isQueryManager',
            deephaven__enterprise_dot_proto_dot_acl__pb2.UnaryUserRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_acl__pb2.BooleanResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def isSchemaManager(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.AclApi/isSchemaManager',
            deephaven__enterprise_dot_proto_dot_acl__pb2.UnaryUserRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_acl__pb2.BooleanResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getUsersForGroup(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.AclApi/getUsersForGroup',
            deephaven__enterprise_dot_proto_dot_acl__pb2.UnaryGroupRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_acl__pb2.UserOrGroupList.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getGroupsForUser(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.AclApi/getGroupsForUser',
            deephaven__enterprise_dot_proto_dot_acl__pb2.UnaryUserRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_acl__pb2.UserOrGroupList.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getAllUsers(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.AclApi/getAllUsers',
            deephaven__enterprise_dot_proto_dot_acl__pb2.GetAllUsersRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_acl__pb2.UserOrGroupList.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getAllGroups(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/io.deephaven.proto.controller.grpc.AclApi/getAllGroups',
            deephaven__enterprise_dot_proto_dot_acl__pb2.GetAllGroupsRequest.SerializeToString,
            deephaven__enterprise_dot_proto_dot_acl__pb2.UserOrGroupList.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



from __future__ import annotations
from enum import Enum

import json
from typing import Union, Optional, Iterable, Iterator, List, Dict, Callable, Any, Tuple

import logging
import threading
import time
import os
import concurrent.futures

import grpc

import deephaven_enterprise.client.util
import deephaven_enterprise.proto.auth_pb2
import deephaven_enterprise.proto.controller_service_pb2_grpc
import deephaven_enterprise.proto.controller_common_pb2
import deephaven_enterprise.proto.controller_pb2
import deephaven_enterprise.proto.persistent_query_pb2
from deephaven_enterprise.client.generate_scheduling import GenerateScheduling

from deephaven_enterprise.client.auth import AuthClient

class SubState(Enum):
    NOT_SUBSCRIBED = 1
    SUBSCRIBING = 2
    SUBSCRIBED = 3

class RefreshThread(threading.Thread):
    def __init__(self,
                 controller_client: "ControllerClient",
                 period_millis: int = 10000):
        """
        This thread pings the controller every 10 seconds (default) to ensure that our cookie is valid and that we stay authenticated.
        :param controller_client:the controller client
        :param period_millis:how often to refresh in milliseconds
        """
        threading.Thread.__init__(self)
        self.stop = False
        self.daemon = True
        self.controller_client = controller_client
        self.period_millis = period_millis

    def run(self):
        while not self.stop and self.controller_client.cookie is not None:
            time.sleep(self.period_millis / 1000)
            try:
                self.controller_client.ping()
            except grpc.RpcError as e:
                if self.controller_client._should_retry_auth(e):
                    self.controller_client.log.info(
                        f'{self.controller_client.me}: Retryable Authentication Error in ping: {str(e)}')
                    self.controller_client._reauthenticate(
                        self.controller_client.rpc_timeout_secs)
                    continue
                if self.controller_client._should_retry(e):
                    self.controller_client.log.info(
                        f'{self.controller_client.me}: Retryable gRPC Error in ping: {str(e)}')
                    continue
                self.controller_client.log.error(f'{self.controller_client.me}: {e}')
                raise e


class ResponseThread(threading.Thread):
    def __init__(self,
                 controller_client: "ControllerClient",
                 iterator: Iterator[deephaven_enterprise.proto.controller_pb2.SubscribeResponse]):
        """
        This thread processes responses from the controller client's subscription method and calls the client's process method.

        :param controller_client: the controller client
        :param iterator: the result of the subscription
        """
        threading.Thread.__init__(self)
        self.daemon = True
        self.controller_client = controller_client
        self.iterator = iterator

    def run(self):
        try:
            for x in self.iterator:
                if self.controller_client.cookie is None:
                    return
                self.controller_client._process_one_event(x)
            self.controller_client.log.info(
                f'{self.controller_client.me}: Controller subscription completed.')
        except grpc.RpcError as e:
            if self.controller_client.cookie is None:
                # If we are logged out, we should not bother
                return
            if not self.controller_client._should_retry(e) and not self.controller_client._should_retry_auth(e):
                self.controller_client.log.error(f'{self.controller_client.me}: {e}')
                raise e
            self.controller_client.log.info(
                f'{self.controller_client.me}: Retryable Error in controller subscription handling: {str(e)}')

        # After our subscription ends, we should re-establish a subscription, unless we are closed
        # (meaning there is no cookie anymore)
        if self.controller_client.cookie is None:
            return

        self.controller_client.log.info(
            f'{self.controller_client.me}: restablishing subscription.')
        self.controller_client.subscribe()


class ControllerClient:
    """
    The ControllerClient connects to the Deephaven PersistentQueryController process.

    You may subscribe to the state of Persistent Queries as well as create and modify them.

    This class operates on the deephaven_enterprise.proto.persistent_query_pb2 structures.
    """

    rpc_timeout_secs: int
    subscription_timeout_secs: int
    query_map: Dict[int,
                    deephaven_enterprise.proto.persistent_query_pb2.PersistentQueryInfoMessage]
    query_map_version: int  # increases when query_map changes
    query_map_condition: threading.Condition  # protects and signals updates to query_map and query_map_version
    effective_user: str
    server_config: deephaven_enterprise.proto.controller_pb2.ControllerConfigurationMessage
    refresh_thread: RefreshThread
    response_thread: ResponseThread
    sub_state_cond: threading.Condition
    sub_state: SubState
    channel: grpc.Channel
    opened_channel: bool
    auth_client: AuthClient
    log: logging.Logger

    def __init__(self, host: str = None, port: int = None, rpc_timeout_seconds: int = 120, channel: grpc.Channel = None, client_name = None):
        """
        Connect to the persistent query controller.

        :param host: the host to connect to, requires port
        :param port: the port to connect to, requires host
        :param rpc_timeout_seconds: the timeout period in seconds for RPCs. Defaults to 120 if not provided.
        :param channel: an already configured gRPC channel (exclusive with host and port), it will not be closed by this client
        """
        self.cookie = None
        self.rpc_timeout_secs = rpc_timeout_seconds
        self.subscription_timeout_secs = rpc_timeout_seconds
        self.query_map = {}
        self.query_map_condition = threading.Condition()
        self.query_map_version = 0
        self.effective_user = None
        self.server_config = None
        self.refresh_thread = None
        self.sub_state_cond = threading.Condition()
        self.sub_state = SubState.NOT_SUBSCRIBED
        self.response_thread = None
        self.auth_client = None
        self.log = logging.getLogger("deephaven_enterprise.controller")
        (self.channel, self.me) = deephaven_enterprise.client.util.get_grpc_channel(
            channel, host, port)
        if client_name is not None:
            self.me += '_' + client_name
        self.opened_channel = channel is None
        self.clientID = deephaven_enterprise.client.util.make_client_id(
            self.me)

        self.stub = deephaven_enterprise.proto.controller_service_pb2_grpc.ControllerApiStub(
            self.channel)
        pingreq = deephaven_enterprise.proto.controller_common_pb2.PingRequest()
        self.stub.ping(pingreq, timeout=self.rpc_timeout_secs,
                       wait_for_ready=True)

    def set_auth_client(self, auth_client: AuthClient):
        """
        Authenticate using the given authentication client, which is stored as a local variable.  If a controller
        operation fails with a gRPC unauthenticated exception; then we attempt to use the authentication client to
        create a new controller session.  If the authentication client cannot produce tokens, then we cannot proceed.
        """
        self.auth_client = auth_client
        self._reauthenticate(self.rpc_timeout_secs)

    def _should_retry_auth(self, e: grpc.RpcError):
        """
        Is this gRPC Error something that should be retried after attempting to reauthenticate to the controller?
        """
        code = e.code()
        return self.auth_client is not None and code.name == "UNAUTHENTICATED"

    def _should_retry(self, e: grpc.RpcError):
        """
        Is this a transient error that we should retry?
        """
        code = e.code()
        if code.name == "UNAVAILABLE":
            return True
        if code.name == "INTERNAL":
            return "RST_STREAM" in e.details()
        return False

    def _reauthenticate(self, timeout: float):
        """
        Request a new token from our authentication client, and then uses it to authenticate to the controller
        """
        if self.auth_client is None:
            raise Exception("Cannot reauthenticate without an auth_client.")
        deadline = time.time() + timeout
        token = self.auth_client.get_token(
            "PersistentQueryController", timeout)
        timeout = deadline - time.time()
        self.authenticate(token, timeout)

    def authenticate(self, token: deephaven_enterprise.proto.auth_pb2.Token, timeout: float = None) -> None:
        """
        Authenticate to the controller using a token obtained from the AuthClient get_token method.

        :param token: the token to use for authentication, must have a service of "PersistentQueryController"
        """
        auth_request: deephaven_enterprise.proto.controller_pb2.AuthenticationRequest = deephaven_enterprise.proto.controller_pb2.AuthenticationRequest()
        auth_request.token.CopyFrom(token)
        auth_request.clientId.CopyFrom(self.clientID)
        auth_request.getConfiguration = True
        use_timeout: float = timeout if timeout is not None else self.rpc_timeout_secs
        auth_response: deephaven_enterprise.proto.controller_pb2.AuthenticationResponse = self.stub.authenticate(
            auth_request, timeout=use_timeout, wait_for_ready=True)
        if not auth_response.authenticated:
            msg = "Could not authenticate to controller."
            self.log.error(f'{self.me}: {msg}')
            raise Exception(msg)
        self.cookie = auth_response.cookie
        if self.refresh_thread is not None:
            self.refresh_thread.stop = True
        self.refresh_thread = RefreshThread(self)
        self.refresh_thread.start()
        self.effective_user = token.user_context.effectiveUser
        if auth_response.config:
            self.server_config = auth_response.config
        else:
            config_request = deephaven_enterprise.proto.controller_pb2.GetConfigurationRequest()
            config_request.cookie = self.cookie
            config_response = self.stub.getConfiguration(config_request,
                                                         timeout=self.rpc_timeout_secs, wait_for_ready=True)
            self.server_config = config_response.config

    def close(self) -> None:
        """
        Invalidate the clients cookie so that further operations do not take place with this client.
        """
        if self.cookie is not None:
            ivcr = deephaven_enterprise.proto.auth_pb2.InvalidateCookieRequest()
            ivcr.cookie = self.cookie
            self.cookie = None
            self.stub.invalidateCookie(
                ivcr, timeout=self.rpc_timeout_secs, wait_for_ready=True)
        if self.opened_channel:
            self.channel.close()

    def _do_subscription(self):
        if self.cookie is None:
            raise Exception("Should authenticate first.")

        self.log.info(f'{self.me}: Starting subscription.')
        subscription_request = deephaven_enterprise.proto.controller_pb2.SubscribeRequest()

        # We have no timeout, because the subscription must last "forever", we use a future executor to handle the
        # actual timeout of the gRPC call
        def _grpc_call():
            subscription = self.stub.subscribe(
                subscription_request, wait_for_ready=True, timeout=None)
            first_value = next(subscription)
            return (subscription, first_value)

        return self._req_with_authentication_retry_no_timeout(subscription_request, _grpc_call)

    def subscribe(self) -> None:
        """
        Subscribe to persistent query state, and wait for the initial query state snapshot to be populated.
        A successful call to authenticate should have happened before this call.

        After the subscription is complete, you may call the map method to retrieve the complete map or the get method
        to fetch a specific query by serial number.
        """

        snap_tstart = time.time()
        snap_nevents = 0
        with self.sub_state_cond:
            self.sub_state = SubState.SUBSCRIBING

        executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
        result_future: concurrent.futures.Future = executor.submit(
            self._do_subscription)
        snap_totsize = 0
        (subscription, first) = result_future.result(
            timeout=self.rpc_timeout_secs)
        snap_nevents += 1
        last: bool = self._process_one_event(first)
        snap_totsize += first.ByteSize()
        while not last:
            event = next(subscription)
            last = self._process_one_event(event)
            snap_totsize += event.ByteSize()
            snap_nevents += 1

        snap_tend = time.time()
        # Process response into our map
        self.response_thread = ResponseThread(self, subscription)
        self.response_thread.start()

        with self.sub_state_cond:
            self.sub_state = SubState.SUBSCRIBED
            self.sub_state_cond.notify_all()

        executor.shutdown()
        snap_dt = snap_tend - snap_tstart
        msg = f'Snapshot for initial subscription done, received {snap_nevents} events in {snap_dt:.2f} seconds'
        if snap_dt > 0:
            msg += f' ({snap_nevents/snap_dt:.2f} e/s), total size {snap_totsize:,} bytes'
            if snap_nevents > 0:
                msg += f' (average event size {snap_totsize/snap_nevents:.2f} bytes)'
        self.log.info(msg)

    def map(self) -> Dict[int, deephaven_enterprise.proto.persistent_query_pb2.PersistentQueryInfoMessage]:
        """
        Retrieve a copy of the current persistent query state.
        A successful call to subscribe should have happened before this call.

        :return: a map from serial number to persistent query info
        """
        deadline = time.time() + self.subscription_timeout_secs
        self._maybe_wait_for_subscription_secs(deadline)
        with self.query_map_condition:
            return self.query_map.copy()

    def map_and_version(self) -> Tuple[
            Dict[int, deephaven_enterprise.proto.persistent_query_pb2.PersistentQueryInfoMessage],
            int]:
        """
        Retrieve a copy of the current persistent query state alongside a version number for the overall map state.
        A successful call to subscribe should have happened before this call.
        Note the version number here has nothing to do with persistent query versions, but reflects instead
        a monotonically increasing version number for the current known state of all subscriptions.
        The version number will increase as update messages arriving from the controller make the map change.
        A successful call to subscribe should have happened before this call.

        :return: A tuple of two elements, a map from serial number to persistent query info, and a version number
        """
        deadline = time.time() + self.subscription_timeout_secs
        self._maybe_wait_for_subscription_secs(deadline)
        with self.query_map_condition:
            return (self.query_map.copy(), self.query_map_version)

    def maybe_wait_for_subscription(self, deadline_ns: int) -> None:
        self._maybe_wait_for_subscription_secs(deadline_ns / (1000.0 * 1000.0 * 1000.0))

    def _maybe_wait_for_subscription_secs(self, deadline: Optional[float] = None) -> None:
        with self.sub_state_cond:
            while True:
                if self.sub_state == SubState.NOT_SUBSCRIBED:
                    raise Exception("Should subscribe first.")
                if self.sub_state == SubState.SUBSCRIBED:
                    return
                if self.sub_state == SubState.SUBSCRIBING:
                    if deadline is not None:
                        now = time.time()
                        if now > deadline:
                            msg = "Deadline exceeded waiting for subscription to finish"
                            self.log.error(f'{self.me}: {msg}')
                            raise Exception(msg)
                        self.sub_state_cond.wait(deadline - now)
                    else:
                        self.sub_state_cond.wait()
                    continue
                # Should not happen unless code logic error
                msg = f'Internal error: sub_state={self.sub_state}'
                self.log.error(f'{self.me}: {msg}')
                raise Exception(msg)

    def get_serial_for_name(self, name: str, timeout_seconds: float = 0) -> int:
        """
        Retrieves the serial number for a given name.

        :param name:             a persistent query name
        :param timeout_seconds:  how long to wait for the query to appear if not already present in the subscription map
        """
        if self.response_thread is None:
            raise Exception("Should subscribe first.")
        deadline = time.time() + timeout_seconds
        with self.query_map_condition:
            while True:
                for serial in self.query_map:
                    qi: deephaven_enterprise.proto.persistent_query_pb2.PersistentQueryInfoMessage = self.query_map[
                        serial]
                    if qi.config.name == name:
                        return serial
                remaining = deadline - time.time()
                if remaining < 0:
                    raise Exception("No query found with name " + name)
                self.query_map_condition.wait(remaining)

    def wait_for_change(self, timeout_seconds: float):
        """
        Waits for a change in the query map to occur.
        See also wait_for_change_from_version.
        """
        with self.query_map_condition:
            self.query_map_condition.wait(timeout_seconds)

    def wait_for_change_from_version(self, map_version: int, timeout_seconds: float) -> bool:
        """
        Waits for a new version in the query map to occur

        :map_version: a version number reflecting the current known map version
        :param timeout_seconds: how long to wait in seconds
        :return: True if a newer version exists, False otherwise
        """
        deadline = time.time() + timeout_seconds
        with self.query_map_condition:
            while True:
                if self.query_map_version > map_version:
                    return True
                remaining = deadline - time.time()
                if remaining < 0:
                    return False
                self.query_map_condition.wait(remaining)

    def get(self, serial: int,
            timeout_seconds: float = 0) -> deephaven_enterprise.proto.persistent_query_pb2.PersistentQueryInfoMessage:
        """
        Get a query from the map.  If the query does not exist, throws a KeyError.
        A successful call to subscribe should have happened before this call.

        The timeout_seconds parameter can be specified to wait for the query to exist before failing with KeyError.
        This is useful when you have just created the query, know its serial, but the controller has not yet published
        the state to you.

        :param serial:
        :param timeout_seconds:
        :return: the PersistentQueryInfoMessage associated with the serial number
        """
        deadline = time.time() + timeout_seconds
        self._maybe_wait_for_subscription_secs(deadline)
        tries = 0
        last_error: Optional[KeyError] = None
        with self.query_map_condition:
            while True:
                try:
                    return self.query_map[serial]
                except KeyError as ke:
                    remaining = deadline - time.time()
                    if remaining < 0:
                        raise ke
                self.query_map_condition.wait(remaining)

    @staticmethod
    def status_name(status: deephaven_enterprise.proto.persistent_query_pb2.PersistentQueryStatusEnum):
        """
        Returns the name of the status enum from a PersistentQueryStateMessage.
        """
        return deephaven_enterprise.proto.persistent_query_pb2.PersistentQueryStatusEnum.Name(status)

    @staticmethod
    def is_running(status: deephaven_enterprise.proto.persistent_query_pb2.PersistentQueryStatusEnum) -> bool:
        """
        Is the status from the query info running?

        If not running and not terminal, then the query is in the initialization process.
        """
        return status == deephaven_enterprise.proto.persistent_query_pb2.PQS_RUNNING

    @staticmethod
    def is_completed(status: deephaven_enterprise.proto.persistent_query_pb2.PersistentQueryStatusEnum) -> bool:
        """
        Is the status from the query info Completed?

        If not running and not terminal, then the query is in the initialization process.
        """
        return status == deephaven_enterprise.proto.persistent_query_pb2.PQS_COMPLETED

    @staticmethod
    def is_terminal(status: deephaven_enterprise.proto.persistent_query_pb2.PersistentQueryStatusEnum) -> bool:
        """
        Is the status from the query info terminal?

        If not running and not terminal, then the query is in the initialization process.
        """
        return status == deephaven_enterprise.proto.persistent_query_pb2.PQS_ERROR or \
            status == deephaven_enterprise.proto.persistent_query_pb2.PQS_DISCONNECTED or \
            status == deephaven_enterprise.proto.persistent_query_pb2.PQS_STOPPED or \
            status == deephaven_enterprise.proto.persistent_query_pb2.PQS_FAILED or \
            status == deephaven_enterprise.proto.persistent_query_pb2.PQS_COMPLETED

    @staticmethod
    def is_status_uninitialized(status: deephaven_enterprise.proto.persistent_query_pb2.PersistentQueryStatusEnum) -> bool:
        """
        Is the status from the query info uninitialized?

        This is the case before a query is ever started.
        """
        return status == deephaven_enterprise.proto.persistent_query_pb2.PQS_UNSPECIFIED or \
            status == deephaven_enterprise.proto.persistent_query_pb2.PQS_UNINITIALIZED

    def _req_with_authentication_retry(self, request, timeout, method):
        """
        Perform the requested method with retries.

        :param request: the request that we'll be sending in method, the cookie for authentication is set on each try
        :param timeout: how long in seconds we have to complete the call
        :param method: the method to execute, must take a single parameter which is the number of seconds for the timeout
        """
        deadline = time.time() + timeout
        while True:
            request.cookie = self.cookie
            try:
                return method(timeout)
            except grpc.RpcError as e:
                timeout = deadline - time.time()
                if timeout > 0:
                    if self._should_retry_auth(e):
                        self.log.info(f'{self.me}: Retryable gRPC Authentication Error: {str(e)}')
                        self._reauthenticate(timeout)
                        continue
                    if self._should_retry(e):
                        self.log.info(f'{self.me}: Retryable gRPC Error: {str(e)}')
                        continue

                    deephaven_enterprise.client.util._maybe_reinterpret_and_raise(e)
                self.log.error(f'{self.me}: {e}')
                raise e

    def _req_with_authentication_retry_no_timeout(self, request, method):
        """
        Perform the requested method with retries, no timeout is provided to the method; the authentication call
        uses our standard timeout.  This is intended for use when we have an external thing managing timeouts (like
        for streaming RPCs).

        :param request: the request that we'll be sending in method, the cookie for authentication is set on each try
        :param method: the method to execute
        """
        while True:
            request.cookie = self.cookie
            try:
                return method()
            except grpc.RpcError as e:
                if self._should_retry_auth(e):
                    self.log.info(f'{self.me}: Retryable gRPC Authentication Error: {str(e)}')
                    self._reauthenticate(self.rpc_timeout_secs)
                    continue
                if self._should_retry(e):
                    self.log.info(f'{self.me}: Retryable gRPC Error: {str(e)}')
                    continue
                deephaven_enterprise.client.util._maybe_reinterpret_and_raise(e)
                self.log.error(f'{self.me}: {e}')
                raise e

    def add_query(self,
                  query_config: deephaven_enterprise.proto.persistent_query_pb2.PersistentQueryConfigMessage) -> int:
        """
        Add a persistent query.
        A successful call to authenticate should have happened before this call.

        :param query_config: the configuration of the query to add.
        :return: the serial number of the created query
        """
        if self.cookie is None:
            raise Exception("Should authenticate first.")
        aqreq = deephaven_enterprise.proto.controller_pb2.AddQueryRequest()
        aqreq.config.CopyFrom(query_config)

        aqresp: deephaven_enterprise.proto.controller_pb2.AddQueryResponse = \
            self._req_with_authentication_retry(
                aqreq,
                timeout=self.rpc_timeout_secs,
                method=lambda timeout: self.stub.addQuery(
                    aqreq, timeout, wait_for_ready=True))

        return aqresp.querySerial

    def modify_query(self,
                     query_config: deephaven_enterprise.proto.persistent_query_pb2.PersistentQueryConfigMessage,
                     do_restart: bool) -> None:
        """
        Modify a persistent query.
        A successful call to authenticate should have happened before this call.

        :param query_config: the configuration of the query to modify.
        :do_restart: whether to restart the modified query.
        """
        if self.cookie is None:
            raise Exception("Should authenticate first.")
        mqreq = deephaven_enterprise.proto.controller_pb2.ModifyQueryRequest()
        mqreq.config.CopyFrom(query_config)
        mqreq.doRestart = do_restart

        self._req_with_authentication_retry(
            mqreq,
            timeout=self.rpc_timeout_secs,
            method=lambda timeout: self.stub.modifyQuery(
                mqreq, timeout, wait_for_ready=True))

    def delete_query(self, serial: int) -> None:
        """
        Delete a query.
        A successful call to authenticate should have happened before this call.

        :param serial: the serial number to delete
        """
        if self.cookie is None:
            raise Exception("Should authenticate first.")
        remove_request: deephaven_enterprise.proto.controller_pb2.RemoveQueryRequest = \
            deephaven_enterprise.proto.controller_pb2.RemoveQueryRequest()
        remove_request.serial = serial

        self._req_with_authentication_retry(remove_request, self.rpc_timeout_secs,
                                            method=lambda timeout: self.stub.removeQuery(remove_request, timeout,
                                                                                         wait_for_ready=True))

    def restart_query(self, serials: Union[Iterable[int], int], timeout_seconds: int = None) -> None:
        """
        Restart one or more queries.
        A successful call to authenticate should have happened before this call.

        :param serials: a query serial number, or an iterable of serial numbers
        """
        if self.cookie is None:
            raise Exception("Should authenticate first.")
        restartreq = deephaven_enterprise.proto.controller_pb2.RestartQueryRequest()
        if isinstance(serials, int):
            restartreq.serials.append(serials)
        else:
            for serial in serials:
                restartreq.serials.append(serial)
        use_timeout: int = timeout_seconds if timeout_seconds is not None else self.rpc_timeout_secs
        self._req_with_authentication_retry(restartreq, use_timeout,
                                            lambda timeout: self.stub.restartQuery(restartreq, timeout, wait_for_ready=True))

    def start_and_wait(self, serial: int, timeout_seconds: int = 120) -> None:
        """
        Start the given query serial number; wait for the query to becoming running (or Completed).
        If the query fails, then raises an exception.

        :param serial: the serial to start
        :param timeout_seconds: how long to wait for the query to become running.
        """
        deadline = time.time() + timeout_seconds

        info = self.map()[serial]
        startLastUpdateNanos = None
        if info.state is not None:
            startLastUpdateNanos = info.state.lastUpdateNanos

        self.restart_query(serial, timeout_seconds=timeout_seconds)

        def is_restarted(maparg: Dict[int, deephaven_enterprise.proto.persistent_query_pb2.PersistentQueryInfoMessage],
                         serial=serial) -> \
                Tuple[bool, deephaven_enterprise.proto.persistent_query_pb2.PersistentQueryInfoMessag]:
            info = maparg[serial]
            if info.state is None:
                return (False, info)
            state = info.state
            if startLastUpdateNanos is not None and state.lastUpdateNanos == startLastUpdateNanos:
                return (False, info)
            status = info.state.status
            if self.is_running(status) or self.is_terminal(status):
                return (True, info)
            return (False, info)

        (ok, info) = self._wait_for_map_predicate(is_restarted, deadline - time.time())
        if ok:
            return

        state = info.state if info is not None else None

        raise Exception("Query (" + str(serial) + ", " +
                        info.config.name + ") is not stopped: " + str(state))

    def stop_query(self, serials: Union[Iterable[int], int], timeout_seconds: int = None) -> None:
        """
        Stop one or more queries.
        A successful call to authenticate should have happened before this call.

        :param serials: a query serial number, or an iterable of serial numbers
        """
        if self.cookie is None:
            raise Exception("Should authenticate first.")
        stopReq = deephaven_enterprise.proto.controller_pb2.StopQueryRequest()
        if isinstance(serials, int):
            stopReq.serials.append(serials)
        else:
            for serial in serials:
                stopReq.serials.append(serial)
        use_timeout: int = timeout_seconds if timeout_seconds is not None else self.rpc_timeout_secs

        self._req_with_authentication_retry(stopReq, use_timeout,
                                            lambda timeout: self.stub.stopQuery(stopReq, timeout, wait_for_ready=True))

    def stop_and_wait(self, serial: int, timeout_seconds: int = 120) -> None:
        """
        Stop the given query serial number; wait for the query to stop (be terminal).
        If the query does not stop in the given time, raise an exception.

        :param serial: the serial to start
        :param timeout_seconds: how long, to wait for the query to become running.
        """
        deadline = time.time() + timeout_seconds

        self.stop_query(serial, timeout_seconds=timeout_seconds)

        def is_stopped(maparg: Dict[int, deephaven_enterprise.proto.persistent_query_pb2.PersistentQueryInfoMessage],
                       serial=serial) -> \
                Tuple[bool, deephaven_enterprise.proto.persistent_query_pb2.PersistentQueryInfoMessag]:
            info = maparg[serial]
            state = info.state
            if state is None:
                return (True, info)
            status = info.state.status
            if self.is_terminal(status) or self.is_status_uninitialized(status):
                return (True, info)
            return (False, info)

        (ok, info) = self._wait_for_map_predicate(is_stopped, deadline - time.time())
        if ok:
            return
        state = None if info is None else info.state

        msg = f'Query ({str(serial)}, {info.config.name}) is not stopped: {str(state)}'
        self.log.error('{self.me}: {msg}')
        raise Exception(msg)

    def wait_for_state(self, state_function: Callable[[], (bool, Any)], timeout_seconds: float = None) -> (bool, Any):
        """
        Wait for a particular state transition in the controller map to occur.

        :param state_function: a function that returns a tuple.  The first element is True if the state transition
        occurred and False otherwise.  The second element is the current state which is then returned to the user.
        :param timeout_seconds: how long, to wait for the state transition to occur
        :return: a tuple.  The first element is True if the state_function returned True; or False if the timeout
        expired.  The second element is the return value of the state_function or None if the state_function returned False.
        """
        if timeout_seconds is None:
            timeout_seconds = self.rpc_timeout_secs
        deadline = time.time() + timeout_seconds
        remaining: int = 0  # always check at least once
        current_state = None
        while True:
            # Don't hold our lock/condition while invoking the callback.
            current_state = state_function()
            if current_state[0]:
                return current_state
            remaining: int = deadline - time.time()
            if remaining < 0:
                return current_state
            with self.query_map_condition:
                self.query_map_condition.wait(remaining)

    def _wait_for_map_predicate(self,
                                map_predicate: Callable[[Dict[int, deephaven_enterprise.proto.persistent_query_pb2.PersistentQueryInfoMessage]],
                                                        (bool, Any)],
                                timeout_seconds: float = None) -> bool:
        # map_predicate should be fast to evaluate, as it is invoked holding our query_map_condition lock.
        deadline: Optional[float] = None
        if timeout_seconds is not None:
            deadline = time.time() + timeout_seconds
        self._maybe_wait_for_subscription_secs(deadline)
        with self.query_map_condition:
            while True:
                found, state = map_predicate(self.query_map)
                if found:
                    return (True, state)
                if deadline is None:
                    self.query_map_condition.wait()
                else:
                    remaining = deadline - time.time()
                    if remaining < 0:
                        return (False, None)
                    self.query_map_condition.wait(remaining)

    def make_temporary_config(self, name: str, heap_size_gb: float,
                              server: str = None,
                              extra_jvm_args: List[str] = None,
                              extra_environment_vars: List[str] = None,
                              engine: str = "DeephavenCommunity",
                              auto_delete_timeout: Optional[int] = 600,
                              admin_groups: List[str] = None,
                              viewer_groups: List[str] = None) -> deephaven_enterprise.proto.persistent_query_pb2.PersistentQueryConfigMessage:
        """
        Create a configuration suitable for use as a temporary InteractiveConsole query.  The worker uses the default
        DeephavenCommunity engine.  This kind of query enables a client to use the controller to create workers for
        general use, in the same way Enterprise clients would have used the dispatcher.  For options that are not
        represented in the arguments, the returned PersistentQueryConfigMessage can be modified before adding it to
        the controller.

        :param name: the name of the persistent query.  Defaults to None, which means a name based on the current time is used
        :param heap_size_gb: the heap size of the worker
        :param server: the server to connect to. Defaults to None, which means the first available server
        :param extra_jvm_args: extra JVM arguments for starting the worker. Defaults to None.
        :param extra_environment_vars: extra Environment variables for the worker. Defaults to None.
        :param engine: which engine (worker kind) to use for the backend worker. Defaults to None, which means
               "DeephavenCommunity"
        :param auto_delete_timeout: after how many seconds should the query be automatically deleted after inactivity.
               Defaults to ten minutes.  If none, auto-delete is disabled.  If zero, the query is deleted immediately after a
               client connection is lost
        :param admin_groups: list of groups that may administer the query.  Defaults to None, which means only the
               current user may administer the query.
        :param viewer_groups: list of groups that may view the query.  Defaults to None, which means only the current
               user may view the query.
        :return: a configuration suitable for passing to add_query.
        """
        if extra_jvm_args is None:
            extra_jvm_args = []
        if extra_environment_vars is None:
            extra_environment_vars = []
        if self.effective_user is None:
            raise Exception(
                "Expect to be authenticated when creating temporary query configuration")
        config: deephaven_enterprise.proto.persistent_query_pb2.PersistentQueryConfigMessage = \
            deephaven_enterprise.proto.persistent_query_pb2.PersistentQueryConfigMessage()
        # Java Long.MIN_VALUE allows the controller to assign a serial number
        config.serial = -2 ** 63
        config.name = name
        config.version = 1
        config.owner = self.effective_user
        config.enabled = True
        config.heapSizeGb = heap_size_gb
        config.bufferPoolToHeapRatio = 0.25
        config.detailedGCLoggingEnabled = True

        if extra_jvm_args is not None:
            config.extraJvmArguments.extend(extra_jvm_args)
        if extra_environment_vars is not None:
            config.extraEnvironmentVariables.extend(extra_environment_vars)

        # config.classPathAdditions
        if server is not None:
            config.serverName = server
        else:
            config.serverName = self.server_config.dbServers[0].name

        if admin_groups is not None:
            config.adminGroups.extend(admin_groups)
        if viewer_groups is not None:
            config.viewerGroups.extend(viewer_groups)
        config.restartUsers = deephaven_enterprise.proto.persistent_query_pb2.RU_ADMIN

        config.scriptCode = ""

        config.scriptLanguage = "Python"
        config.configurationType = "InteractiveConsole"

        if auto_delete_timeout is not None:
            auto_delete_value = True;
            tsf = {
                "TerminationDelay": auto_delete_timeout * 1000,
            }
            config.typeSpecificFieldsJson = self.__encode_type_specific_fields(
                tsf)
        else:
            auto_delete_value = False;

        scheduling = GenerateScheduling.generate_temporary_scheduler(expiration_time_minutes=2880,
            queue_name="InteractiveConsoleTemporaryQueue",
            auto_delete=auto_delete_value)

        config.scheduling.extend(scheduling)

        # How long to wait for the query to initialize, without an initialization script the default of 60 seconds
        # should be safe on most installations.
        initialize_timeout_seconds: float = 60.0
        config.timeoutNanos = int(initialize_timeout_seconds * 1_000_000_000)

        config.jvmProfile = "Default"

        # config.kubernetesControl = None
        config.workerKind = engine

        return config

    @staticmethod
    def generate_disabled_scheduler():
        """
        Generates a scheduler array for a PQ that has scheduling disabled.
        """
        return GenerateScheduling.generate_disabled_scheduler()

    @staticmethod
    def __encode_type_specific_fields(tsf: Dict) -> str:
        """
        Encodes type specific fields from a Python dictionary into the JSON object that the controller expects.
        :param tsf: a Python dictionary with type specific fields
        :return: a JSON encoded string suitable for the controller
        """
        if tsf is None or len(tsf) == 0:
            return ""

        encoded = {}
        for (k, v) in tsf.items():
            if v is None:
                continue
            if isinstance(v, int):
                encoded[k] = {"type": "long", "value": str(v)}
            elif isinstance(v, bool):
                encoded[k] = {"type": "boolean",
                              "value": "true" if v else "false"}
            elif isinstance(v, str):
                encoded[k] = {"type": "string", "value": v}
            elif isinstance(v, float):
                encoded[k] = {"type": "double", "value": v}

        return json.dumps(encoded)

    def ping(self):
        """
        Ping the controller and refresh our cookie.
        :return: True if the ping was sent, False if we had no cookie
        """
        ping_request = deephaven_enterprise.proto.controller_common_pb2.PingRequest()
        cookie = self.cookie
        if cookie is None:
            return False

        ping_request.cookie = cookie
        self.stub.ping(
            ping_request, timeout=self.rpc_timeout_secs, wait_for_ready=True)
        return True

    def _process_one_event(self, event: deephaven_enterprise.proto.controller_pb2.SubscribeResponse):
        """
        Process a controller event.
        :param event: the event to process
        :return: True if this is the end of the initial subscription batch
        """
        batch_end: bool = event.event == deephaven_enterprise.proto.controller_pb2.SubscriptionEvent.SE_BATCH_END
        if event.event == deephaven_enterprise.proto.controller_pb2.SubscriptionEvent.SE_PUT or batch_end:
            serial = event.queryInfo.config.serial
            with self.query_map_condition:
                self.query_map[serial] = event.queryInfo
                self.query_map_version += 1
                self.query_map_condition.notify_all()
        elif event.event == deephaven_enterprise.proto.controller_pb2.SubscriptionEvent.SE_REMOVE:
            serial = event.querySerial
            if serial in self.query_map:
                with self.query_map_condition:
                    del self.query_map[serial]
                    self.query_map_version += 1
                    self.query_map_condition.notify_all()
            else:
                raise Exception("Unknown query: " + str(event))
        elif event.event == deephaven_enterprise.proto.controller_pb2.SubscriptionEvent.SE_CONFIG_UPDATE:
            self.server_config = event.config
        else:
            msg = f'Unexpected event: {str(event)}'
            self.log.error(f'{self.me}: msg')
            raise Exception(msg)
        return batch_end



#
#  Copyright (c) 2016-2021 Deephaven Data Labs and Patent Pending
#
import pathlib
from setuptools import setup

import os, re

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

if "DH_FROM_BUILD_DIR" in os.environ and os.environ["DH_FROM_BUILD_DIR"]:
    gradle_properties = (HERE / "../../../../gradle.properties")
else:
    # These work from the source directory
    gradle_properties = (HERE / "../../../../../../gradle.properties")

props = dict()
for line in open(gradle_properties, "r").readlines():
    line = line.strip().split("#", 2)[0].strip()
    if line == "":
        continue
    kv = line.split("=")
    if kv[0] in ["dhcVersion", "versionSource"]:
        props[kv[0]] = kv[1]

if "versionSource" not in props:
    raise Exception("Could not determine versionSource from %s" % gradle_properties)
versionSource = props["versionSource"]

if os.environ["DH_FROM_BUILD_DIR"]:
    versionFile = (HERE / ("../../../../gradle/versions/%s" % versionSource))
else:
    # These work from the source directory
    versionFile = (HERE / ("../../../../../../gradle/versions/%s" % versionSource))

client_dir = "deephaven_enterprise/client"
proto_dir = "deephaven_enterprise/proto"

VERSION = versionFile.read_text().strip()

VERSION_BITS = VERSION.split('.')
if len(VERSION_BITS[2]) > 3:
    VERSION_BITS[2] = VERSION_BITS[2][0:12] + '+' + VERSION_BITS[2][12:]

VERSION = '.'.join(VERSION_BITS)

dhcVersion = os.environ["DHC_VERSION"] if "DHC_VERSION" in os.environ else None

if dhcVersion is None:
    dhcVersion = props['dhcVersion']
if dhcVersion is None:
    raise Exception("Could not determine dhcVersion from %s" % gradle_properties)

setup(
    name='deephaven-coreplus-client',
    version=VERSION,
    description='The Deephaven Enterprise Core+ Python Client',
    long_description=README,
    long_description_content_type="text/markdown",
    packages=["deephaven_enterprise.client", "deephaven_enterprise.proto"],
    package_dir={"deephaven_enterprise.client" : client_dir, "deephaven_enterprise.proto" : proto_dir},
    url='https://deephaven.io/',
    license='Deephaven Enterprise License Agreement',
    author='Deephaven Data Labs',
    author_email='python@deephaven.io',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: Other/Proprietary License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    python_requires='>=3.8',
    install_requires=['pydeephaven==' + dhcVersion, 'pycryptodomex>=3.19.0', 'pycryptodome>=3.19.0', 'grpcio>=1.63.0', 'urllib3', 'requests']
)
